﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Reflection;
using System.Reflection.Emit;

namespace Core
{
    public static class Make
    {
        public static Action<T> Action<T>(Action<T> action) { return action; }

        public static Func<TResult> Func<TResult>(Func<TResult> fun) { return fun;}
        public static Func<T, TResult> Func<T, TResult>(Func<T, TResult> fun) { return fun; }
        public static Func<T1, T2, TResult> Func<T1, T2, TResult>(Func<T1, T2, TResult> fun) { return fun; }
        public static Func<T1, T2, T3, TResult> Func<T1, T2, T3, TResult>(Func<T1, T2, T3, TResult> fun) { return fun; }


        private static Dictionary<Type, Type> _implementations = new Dictionary<Type, Type>();

        public static T Instance<T>(object methods) where T : class //T must be an interface
        {
            Type t = typeof(T);
            if (!t.IsInterface)
            {
                throw new ArgumentException("T must be an interface.");
            }

            Type concrete;
            lock (_implementations)
            {
                if (!_implementations.ContainsKey(t))
                {
                    _implementations[t] = Implement(t, methods);
                }

                concrete = _implementations[t];
            }

            return (T)Activator.CreateInstance(concrete, methods);
        }

        private static Type Implement(Type targetInterface, object methods)
        {
            string baseName = targetInterface.Name + "-Dynamic";
            AssemblyName name = new AssemblyName { Name = baseName };
            AssemblyBuilder assemblyBuilder = AppDomain.CurrentDomain.DefineDynamicAssembly(name, AssemblyBuilderAccess.RunAndSave);
            ModuleBuilder moduleBuilder = assemblyBuilder.DefineDynamicModule(baseName + ".dll");
            TypeBuilder typeBuilder =
                moduleBuilder.DefineType(
                    baseName + ".Impl",
                    TypeAttributes.Public | TypeAttributes.Sealed | TypeAttributes.Class,
                    typeof(Record),
                    new Type[] { targetInterface });
            //emit ctor
            ConstructorBuilder ctorBuilder =
                typeBuilder.DefineConstructor(
                    MethodAttributes.Public,
                    CallingConventions.Standard,
                    new Type[] { typeof(object) });
            ILGenerator ctorBody = ctorBuilder.GetILGenerator();
            ctorBody.Emit(OpCodes.Ldarg_0);
            ctorBody.Emit(OpCodes.Ldarg_1);
            ctorBody.Emit(OpCodes.Call, typeBuilder.BaseType.GetConstructor(new Type[] { typeof(object) }));
            ctorBody.Emit(OpCodes.Ret);

            //emit methods
            Record impls = new Record(methods);
            foreach (MethodInfo method in targetInterface.GetMethods())
            {
                MethodBuilder methodBuilder =
                    typeBuilder.DefineMethod(
                        method.Name,
                        MethodAttributes.Public | MethodAttributes.Final | MethodAttributes.Virtual | MethodAttributes.HideBySig | MethodAttributes.NewSlot,
                        method.ReturnType,
                        method.GetParameters().Select(p => p.ParameterType).ToArray());

                ILGenerator methBody = methodBuilder.GetILGenerator();
                if (impls.ContainsProperty(method.Name))
                {
                    Type implType = impls[method.Name].GetType();
                    Type[] implGenericArgs = implType.GetGenericArguments();

                    if (!implType.Name.StartsWith("Func`"))
                    {
                        throw new ArgumentException("Methods must be implemented by Funcs");
                    }                    

                    methBody.Emit(OpCodes.Ldarg_0);
                    methBody.Emit(OpCodes.Ldstr, method.Name);
                    methBody.Emit(OpCodes.Call, typeof(Record).GetMethod("Get").MakeGenericMethod(implType));
                    for (int arg = 1; arg < implGenericArgs.Length; arg++)
                    {
                        methBody.Emit(OpCodes.Ldarg, arg);
                    }

                    methBody.Emit(OpCodes.Callvirt, implType.GetMethod("Invoke"));
                    methBody.Emit(OpCodes.Ret);                    
                } else
                {
                    //throw new NotImplementedException();
                    methBody.Emit(OpCodes.Newobj, typeof(NotImplementedException).GetConstructor(System.Type.EmptyTypes));
                    methBody.Emit(OpCodes.Throw);
                }
            }

            return typeBuilder.CreateType();
        }
    }
}
